#
# Copyright (C) 1993 Brian Kahn, Steve O'Keefe, Bob Noel
# Copyright (C) 1993 All Rights Reserved
# Copyright (C) 1993 Thu May 13 1993
#
# Complete copyright is held by each of the authors
# Permission to distribute and use for educational purposes
#
# 

# User defined procedures

proc arcReshape {{tag selected}} {
    # arcs have begin, middle, end
    global c x y
    set arc [itemBase $tag]
    set last [guiArcReadPoint $arc]
    set lastX [lindex $last 0]
    set lastY [lindex $last 1]
    if {[itemType $arc]=="arc"} then {
        guiArcWritePoint $arc $x $y
        eval "$c coords $arc [guiArcAdjust $arc]"
        $c move weight$arc [expr $x-$lastX] [expr $y-$lastY]
    }
}


proc arcSink {} {
    # finish drawing an arc
    global c
    set phantom [$c find withtag phantom]
    set src [$c find withtag arcSrc]
    set sink_at [lrange [$c coords $phantom] 2 3]
    $c delete $phantom
    $c dtag phantom
    $c dtag arcSrc
    if {($phantom=="") || ($src=="") || ($sink_at=="")} then {
        error "Internal error in arcSink: $phantom, $src, $sink_at."
    }
    # do nothing if there is no sink
    set sink [itemBase \
        [lindex [eval "$c find overlapping $sink_at $sink_at"] 0]]
    if {$sink!=""} then {
        eval "new_arc $src $sink"
    }
}


proc arcSrc {start} {
    # start drawing an arc
    global c x y
    set src [itemBase $start]
    if {$src==""} {error "arcSrc: nothing selected."}
    $c addtag arcSrc withtag $src
    eval "$c create line [itemCenter $src] $x $y \
        -arrow last -tag phantom"
}


proc arcStretch {} {
    # move phantom arc
    global c x y
    if {[set phantom [$c find withtag phantom]] != ""} then {
        eval "$c coords $phantom [lrange [$c coords $phantom] 0 1] $x $y"
    }
}


proc between {x1 y1 x2 y2} {
    # point between two others
    return [list [expr {($x1 + $x2) / 2}] [expr {($y1 + $y2) / 2}]]
}


proc betweenItems {item1 item2} {
    # point between two items
    return [eval "between [itemCenter $item1] [itemCenter $item2]"]
}


proc errorMessage {string} {
    # display warning
    global warningMessage
    set warningMessage $string
}


proc fire {} {
    # fire all enabled transitions
    global c statusMessage warningMessage
    global auto_conflict_mode fireStatus
    set statusMessage "Firing..."
    # if marking was modified, save state
    global marking_modified
    if {$marking_modified} {
    	stateAdd
    	set marking_modified 0
    }    	
    # fire!
    fireSetLoad
    # catch max tokens error
    if {[catch guiConflictPlaceSet]} {
    	set statusMessage ""
    	set warningMessage "The net is unfirable - max tokens exceeded."
    	return
    }

    if {[set trans [guiFire]]==""} {
    	# conflicts, in manual mode
        set warningMessage "Conflicts must be resolved"
    } else {
    	# successfully fired
        set warningMessage ""
    	if {$trans!=0} {
            stateAdd
    	}
    	if {$trans==1} {
    	    set statusMessage "Fired 1 transition."
    	} else {
       	    set statusMessage "Fired $trans transitions."
    	}
    	# remove all holds (hold is a temp inhibit)
    	foreach t [array names fireStatus] {
    	    if {$fireStatus($t)=="hold"} {
    	    	set fireStatus($t) fire
    	    	$c itemconfigure hold$t -fill {}
    	    }
    	}
    }
    displayRefresh
}



proc fireSetLoad {} {
    # load the fire set, which is all enabled trans without inhibits
    global fireStatus
    set fireset {}
    foreach t [guiEnabledTransSet] {
    	if {$fireStatus($t)=="fire"} {
    	    lappend fireset $t
    	}
    }
    eval "guiFireTransSet $fireset"
    return $fireset
}


proc fsbox_load {} {
    # handle save from filebox
    fsbox_set_filename
    netLoad
}


proc fsbox_save {} {
    # handle save from filebox
    fsbox_set_filename
    netSave
}


proc fsbox_set_filename {} {
    # set the working filename
    global netname filename fsBox
    set netname $fsBox(path)/$fsBox(name)
    # strip extension if present
    if {! [regsub (.*)\\.net $netname \\1 filename]} {
    	set filename $netname
    }
}


proc graphInvariants {{type {s t}} {file {}}} {
    # do invariants
    global filename statusMessage
    if {$file=={}} {set file $filename}
    foreach t $type {
        case $t in {
    	    {s} {guiGraphInvariants ${file}.sin 1}
    	    {t} {guiGraphInvariants ${file}.tin 2}
    	}
    }
    set statusMessage "Generated invariants in '${file}.${type}in'"
}


proc graphReachable { {file {}} {limit {}} } {
    # generate occurrence graph to FILE
    global filename occurrenceMax statusMessage
    if {$file=={}} {set file $filename}
    if {$limit=={}} {set limit $occurrenceMax}
    guiGraphReachable ${file}.occ $limit
    set statusMessage "Occurrence graph in '${file}.occ'"
}


proc graphCoverage { {file {}} {limit {}} } {
    # generate coverage graph to FILE
    global filename state_current statusMessage state_list occurrenceMax
    if {$file=={}} {set file $filename}
    if {$limit=={}} {set limit $occurrenceMax}
    set state $state_current
    stateSelect 0
    guiGraphCoverage ${file}.cov $limit [lindex $state_list $state]
    stateSelect $state
    set statusMessage "Coverage graph in '${file}.cov'"
}


proc itemBase {item} {
    # base object for item - base is 1st in precedence
    if {[string match {[0-9]*} $item]} {return $item}
    global c
    return [lindex [$c find withtag [itemName $item]] 0]
}


proc itemCenter {tagOrId} {
    # returns center of an item
    global c
    # item must exist
    if {! [itemExists $tagOrId]} then {error "Can't find $tagOrId"}
    set coords [$c coords [itemBase $tagOrId]]
    if {[llength $coords]==2} then {
        return $coords
    } else {
        return [eval "between $coords"]
    }
}


proc itemDup {tag} {
    # Copy, raise, select item
    global c
    set item [itemBase $tag]
    if {$item==""} {error "itemDup: nothing selected."}
    set type [itemType $item]
    set new [eval "new_$type [itemCenter $item]"]
    itemUnhighlight $tag
    itemUnselect $tag
    $c raise $new
    itemHighlight $new
    itemSelect $new
    $c move $new 3 3
    return $new
}


proc itemExists {tag} {
    # return exists boolean
    global c
    if {[$c find withtag $tag] == ""} then {return 0} else {return 1}
}


proc itemMove {{tag selected}} {
    global c x y lastX lastY
    set id [itemBase $tag]
    if {$id!=""} {
    	$c move $tag [expr $x-$lastX] [expr $y-$lastY]
  	eval "guiNodeWriteRect $id [$c coords $id]"
    	foreach arc [guiArcsOfNode $id] {
            eval "$c coords $arc [guiArcAdjust $arc]"
        }
    }
}


proc itemName {tag} {
    #find item name in tag list - shaky
    global c
    set name [lindex [$c gettags $tag] 0]
    if {$name==""} {return $tag} {return $name}
}

proc itemSelect {tag} {
    global c lastSelect
    # work with name, not current
    set name [itemName $tag]
    $c raise $name
    $c addtag selected withtag $name
    set lastSelect $name
}


proc itemType {item} {
    # returns petrinet type of item; place, trans, or arc
    global c
    if {$item!=""} {
        set tags [$c gettags $item]
        if {$tags==""} {return}
        regexp "place|trans|arc|text" $tags match
        return $match
    }
}


proc itemUnselect {{tag all}} {
    # unselect TAG, or no
    global c
    $c dtag [itemName $tag] selected 
}


proc lreverse {list} {
    # reverses a list
    set tmp {}
    foreach i $list {set tmp [concat $i $tmp]}
    return $tmp
}


proc markingAdjust {place {incr 1}} {
    # adjust a place's marking
    global c
    set place [itemBase $place]
    set value [expr [guiPlaceReadMarking $place]+$incr]
    if {$value>=0} {
        markingSet $place $value
    }
}


proc killObject {{tag selected}} {
    # delete the item, and associated arcs
    global c net
    set item [itemBase $tag]
    case [itemType $item] in {
        # remove the logical net entries
        arc {
    	    guiArcDelete $item
        }
        trans {
            foreach arc [guiArcsOfNode $item] {
            	# delete the arc with recursive call
    	    	killObject $arc
            }
    	    global fireStatus
    	    unset fireStatus($item)
    	    guiTransDelete $item
        }
        place {
            foreach arc [guiArcsOfNode $item] {
            	# delete the arc with recursive call
    	    	killObject $arc
            }
    	    guiPlaceDelete $item
        }
    }
    $c delete [itemName $item]
    statePurge
}


proc new_arc {src sink} {
    global c normal_arc_color
    # create a new arc
    # one must be a place, the other a trans
    set srcType [itemType $src]
    set sinkType [itemType $sink]
    set mid [betweenItems $src $sink]
    set xx [lindex $mid 0]
    set yy [lindex $mid 1]
    if {(($srcType=="place") && ($sinkType=="trans")) ||
            (($srcType=="trans") && ($sinkType=="place"))} then {
        # first create arc anywhere, then reposition it
        set arc [mk_arc $xx $yy]
        guiArcCreate $arc $src $sink
        guiArcWritePoint $arc $xx $yy
        eval "$c coords $arc [guiArcAdjust $arc]"
    	weightSet $arc 1
        statePurge
        return arc$arc
    }
}


proc new_place {xx yy {marking 0}} {
    # init
    global c normal_place_color normal_text_color placeX placeY placeCnt
    set item [mk_place [expr $xx-$placeX] [expr $yy-$placeY] \
    	[expr $xx+$placeX] [expr $yy+$placeY]]
    guiPlaceCreate $item
    eval "guiNodeWriteRect $item [$c coords $item]"
    markingSet $item $marking
    $c raise place$item
    guiNodeWriteLabel $item p$placeCnt
    guiNodeWriteLabelOffset $item 1 1
    guiPlaceWriteMaxTokens $item 99
    labelShow $item
    statePurge
    return place$item
}


proc new_placeXY {} {
    global x y
    itemUnselect all
    itemUnbold all
    itemSelect [new_place $x $y]
}


proc new_trans {xx yy} {
    # init
    global c transX transY transCnt
    set item [mk_trans [expr $xx-$transX] [expr $yy-$transY] \
    	[expr $xx+$transX] [expr $yy+$transY]]
    guiTransCreate $item
    eval "guiNodeWriteRect $item [$c coords $item]"
    $c raise trans$item
    guiNodeWriteLabel $item t$transCnt
    guiNodeWriteLabelOffset $item 1 1
    labelShow $item
    statePurge
    return trans$item
}


proc new_transXY {} {
    global x y
    itemUnselect all
    itemUnbold all
    itemSelect [new_trans $x $y]
}


proc netBbox {} {
    # returns a bbox size for net, ignoring exaggerated arc centers
    global c
    # use the true center points for sizing
    foreach a [guiArcSet] {
    	set new [eval "lreplace \[$c coords $a\] 2 3 [guiArcReadPoint $a]"]
    	eval "$c coords $a $new"
    }
    set size [$c bbox all]
    # set the arc centers back
    foreach arc [guiArcSet] {
        eval "$c coords $arc [guiArcAdjust $arc]"
    }
    return $size
}


proc netLoad {{file {}}} {
    # load net from a FILE
    global c filename statusMessage net_modified
    set statusMessage "Loading..."
    netPurge
    if {$file=={}} {
    	set file $filename.net
    }
    if {[catch {guiNetRead $file}]} {
    	errorMessage "File load error"
    	return
    }
    labelRefresh
    # position view of net
    set ww [lrange [$c configure -width] 4 4]
    set wh [lrange [$c configure -height] 4 4]
    set size [netBbox]
    # empty net has no size
    if {$size==""} {return}
    set nw [expr [lindex $size 2]-[lindex $size 0]]
    set nh [expr [lindex $size 3]-[lindex $size 1]]
    # get net x, y in scroll incrs (10)
    set nx [lindex $size 0]
    set ny [lindex $size 1]
    # goto upper left if net is bigger than window
    if {$ww<$nw} {
        $c xview [expr $nx/10]
    } else {
    	$c xview [expr ($nx-($ww/2))/10]
    }
    if {$wh<$nh} {
        $c yview [expr $ny/10]
    } else {
    	$c yview [expr ($ny-($wh/2))/10]
    }
    set statusMessage "loaded at $size"
    # record initial state
    stateSetInitial
    set net_modified 0
    # refresh here will eliminate circle highlights from places
    displayRefresh
}


proc netPrint {} {
    # printing - first attempt
    global c filename statusMessage
    set size [netBbox]
    # empty net has no size
    if {$size==""} {return}
    $c postscript -file ${filename}.ps -colormode mono \
    -pageheight 10.5i -pagewidth 8i -pagex 4i -pagey 5.25i \
    	-x [expr [lindex $size 0]-36] -y [expr [lindex $size 1]-36] \
    	-width [expr [lindex $size 2]+108] -height [expr [lindex $size 3]+108] 
    set statusMessage "Postscript drawing in '${filename}.ps'"
}


proc netPurge {} {
    # clear the net
    global c placeCnt transCnt statusMessage net_modified state_list
    netSaveYN
    guiNetPurge
    # record null initial state
    set state_list {}
    stateSetInitial
    set net_modified 0
    set placeCnt 0
    set transCnt 0
    $c delete all
    set statusMessage "Net erased"
}
    


proc netQuit {} {
    # quit xpetri, with confirmation
    netSaveYN
    exit
}


proc netSaveYN {} {
    # query user about saving net
    global net_modified
    if {$net_modified} {
        if {[YesNoBox "Save the current net?" +100+100]} {
    	    netSave
    	}
    }
}


proc netSave {{file {}}} {
    # save net in FILE
    global filename statusMessage
    set statusMessage "Saving..."
    if {$file=={}} {
    	set file $filename.net
    }
    if {[catch {guiNetWrite $file}]} {
    	errorMessage "File write error"
    	return
    }
    set statusMessage "Saved net in '$file'"
}
    


proc panelPop {{item current}} {
    # prepare and post the panel for ITEM or current
    set type [itemType $item]
    panelPrepare $type $item
    # make it map on top
    wm withdraw [SN ${type}_panel]
    wm deiconify [SN ${type}_panel]
}


proc panelPrepare {type {item {}}} {
    # post modifier panel for item ID
    case $type in {
	{arc} {
	    upvar #0 arc_panel_id id
	    set id [itemBase $item]
	    global c
	    [SN arc_panel_weight] set [guiArcReadArcExpression $id]
	}
	{trans} {
	    upvar #0 trans_panel_id id
	    set id [itemBase $item]
	    global trans_panel_label trans_panel_label_offset
	    [SN trans_panel_label] delete 1.0 end
	    [SN trans_panel_label] insert end [guiNodeReadLabel $id]
	    set trans_panel_label_offset [guiNodeReadLabelOffset $id]
	    # inhibit status
	    global fireStatus trans_panel_inhibit
    	    set trans_panel_inhibit $fireStatus($id)
    	}
	{place} {
	    upvar #0 place_panel_id id
	    set id [itemBase $item]
	    global place_panel_label place_panel_label_offset
	    global place_panel_marking 
	    global place_panel_max_tokens place_panel_min_tokens
	    [SN place_panel_label] delete 1.0 end
	    [SN place_panel_label] insert end [guiNodeReadLabel $id]
	    set place_panel_label_offset [guiNodeReadLabelOffset $id]
	    set place_panel_marking [guiPlaceReadMarking $id]
	    set place_panel_max_tokens [guiPlaceReadMaxTokens $id]
	    set place_panel_min_tokens [guiPlaceReadMinTokens $id]
	    [SN place_panel_marking] set $place_panel_marking
	    [SN place_panel_max_tokens] set $place_panel_max_tokens
	    [SN place_panel_min_tokens] set $place_panel_min_tokens
	}
	{xpetri} {
	    global arcSplinesteps
	    [SN xpetri_panel_splinesteps] set $arcSplinesteps
    	    global xpetri_panel_occurrence_max occurrenceMax \
    	    	xpetri_panel_save_state_max stateMax
    	    set xpetri_panel_occurrence_max $occurrenceMax
    	    set xpetri_panel_save_state_max $stateMax
    	}
    }
}



proc panelProcess type {
    # process the results of a panel
    global c
    case $type in {
	{arc} {
	    upvar #0 arc_panel_id id
	    global c
	    weightSet $id [[SN arc_panel_weight] get]
	}
	{trans} {
	    upvar #0 trans_panel_id id
	    global trans_panel_label trans_panel_label_offset
	    guiNodeWriteLabel $id [[SN trans_panel_label] get 1.0 end]
	    eval "guiNodeWriteLabelOffset $id $trans_panel_label_offset"
	    labelShow $id
	    # set inhibit/fire status
	    global trans_panel_inhibit
	    statusSet $id $trans_panel_inhibit
	}
	{place} {
	    upvar #0 place_panel_id id
	    global place_panel_label place_panel_label_offset
	    guiNodeWriteLabel $id  [[SN place_panel_label] get 1.0 end]
	    eval "guiNodeWriteLabelOffset $id $place_panel_label_offset"
	    labelShow $id
	    markingSet $id [[SN place_panel_marking] get]
	    guiPlaceWriteMaxTokens $id [[SN place_panel_max_tokens] get]
	    guiPlaceWriteMinTokens $id [[SN place_panel_min_tokens] get]
	}
	{xpetri} {
	    global arcSplinesteps
	    if {$arcSplinesteps!=[[SN xpetri_panel_splinesteps] get]} {
		set arcSplinesteps [[SN xpetri_panel_splinesteps] get]
		foreach arc [guiArcSet] {
		    $c itemconfigure $arc -splinesteps $arcSplinesteps
		}
	    }
    	    global xpetri_panel_occurrence_max occurrenceMax \
    	    	xpetri_panel_save_state_max stateMax
    	    set occurrenceMax $xpetri_panel_occurrence_max
    	    set stateMax $xpetri_panel_save_state_max
    	}
    }
}


proc lremove {a args} {
    # remove all ARGS from list A
    # used by inhibit routines
    upvar $a list
    foreach item $args {
    	set index [lsearch $list $item]
    	if {$index>=0} {
    	    set list [lreplace $list $index $index]
    	}
    }
    return $list
}


# state_list[0] is the initial state
# PREV(1) -> (1)   NEXT(last) -> last
# PREV(0) -> (0)   NEXT(0) -> (1) ??

# firing wipes the rest of history
# setting initial state or modifying state wipes all history
# modifying net sets the initial state


proc stateAdd {} {
    # add a new state to the history list, purging all past state_current
    global history stateMax state_scale_to
    if {!$history} {return}
    global c state_list state_cnt state_current
    # wipe the rest of history
    if {$state_cnt>$state_current+1} {
        statePurge [expr $state_current+1]
    }
    lappend state_list [stateCreate]
    set state_cnt [llength $state_list]
    # remember that initial state is state 0
    if {$state_cnt>($stateMax+1)} {
    	# max states exceeded, remove some states
    	set remove [expr $state_cnt-($stateMax+1)]
    	set state_list [lreplace $state_list 1 $remove]
    	set state_cnt [llength $state_list]
    } else {
        incr state_current
    	[SN state_scale] configure -to $state_current
        [SN state_scale] set $state_current
    	set state_scale_to $state_current
    }
}


proc stateCreate {} {
    # create a state, return it's ID
    # state ids are decreasing negative numbers
    global state_id
    incr state_id -1
    guiStateCreate $state_id
    return $state_id
}    


proc stateInitial {} {
    # select initial marking
    stateSelect 0
}


proc stateNext {} {
    # set net state to next in history list, if any
    # not an error if we can't change state
    global history
    if {!$history} {return}
    global state_list state_cnt state_current
    if {$state_cnt>$state_current+1} {
    	stateSelect [expr $state_current+1]
    }
}


proc stateOne {} {
    # set state one
    statePurge 1
    stateAdd
}


proc statePrev {} {
    # set net state to next in history list, if any
    # not an error if we can't change state
    global history
    if {!$history} {return}
    global state_list state_cnt state_current
    if {$state_current>0} {
    	stateSelect [expr $state_current-1]
    }
}


proc statePurge {{curr 0}} {
    # purge the state history from start or CURR to end, inclusive
    global c state_list state_cnt state_current state_scale_to
    global marking_modified net_modified
    foreach s [lrange $state_list $curr end] {
    	guiStateDelete $s
    }
    set state_list [lrange $state_list 0 [expr $curr-1]]
    # always need a state 0
    if {$curr==0} {    
    	set state_list [stateCreate]
        set marking_modified 0
    	# changes to the net always purge state history
    	# all net modifications are captured here
    	set net_modified 1
    }
    set state_cnt [llength $state_list]
    set state_current [expr $state_cnt-1]
    [SN state_scale] configure -to $state_current
    set state_scale_to $state_current
}


proc stateSelect {n} {
    # goto state N, or state indicated on scale if n<0
    global history statusMessage
    if {!$history} {return}
    global state_list state_cnt state_current
    # There is a bug in Tk3.2 that calls the scale command when it is set
    # this if stmt breaks the loop
    if {$n==$state_current} {return}
    if {$n>=$state_cnt || $n<0} {
    	error "stateSelect: invalid state, failed 0<$n<$state_cnt"
    }
    guiRevertToState [lindex $state_list $n]
    set state_current $n
    set statusMessage "Selected state $state_current"
    displayRefresh
    [SN state_scale] set $n
}
    

proc stateSetInitial {} {
    # set the initial state
    statePurge
}



proc statusAdjust {item {way 1}} {
    # rotate status
    global c fireStatus
    set id [itemBase $item]
    case "$fireStatus($id) $way" in {
    	{"fire 1"} {statusSet $id hold}
    	{"hold 1"} {statusSet $id inhibit}
    	{"inhibit 1"} {statusSet $id fire}
    	{"fire -1"} {statusSet $id inhibit}
    	{"inhibit -1"} {statusSet $id hold}
    	{"hold -1"} {statusSet $id fire}
    	default {error "statusSet: '$fireStatus($id) $way' for '$id'"}
    }
}
    

proc statusSet {id status} {
    # decorate trans box - / means hold, X means inhibit
    global c fireStatus
    case $status in {
	{fire} {
	    $c itemconfigure hold$id -fill {}
	    $c itemconfigure inhibit$id -fill {}
	}
	{hold} {
	    $c itemconfigure hold$id -fill black
	    $c itemconfigure inhibit$id -fill {}
	}
	{inhibit} {
	    $c itemconfigure hold$id -fill black
	    $c itemconfigure inhibit$id -fill black
	}
    	{default} {
    	    error "statusSet: bad status '$status' for '$id'"
    	}
    }
    set fireStatus($id) $status
    displayRefresh
}

proc weightAdjust {arc {incr 1}} {
    # adjust a arc's weight
    global c
    set arc [itemBase $arc]
    set value [expr [guiArcReadArcExpression $arc]+$incr]
    if {$value>=0} {
        weightSet $arc $value
    }
}


proc weightSet {arc weight} {
    # set the weight value for an arc
    global c
    set arc [itemBase $arc]
    guiArcWriteArcExpression $arc $weight
    weightShow $arc $weight
}



proc xySelect {} {
    global c x y
    # select the item at x y
    set item [lindex [$c find overlapping $x $y $x $y] 0]
    if {$item != ""} {
        itemSelect $item
        return $item
    }
}


proc xyScroll {var bar total win first last} {
  # record values in var and invoke scrollbar
  global c
  upvar #0 $var record
  set incr [lindex [$c configure -scrollincrement] 4]
  set record(first) [expr $first*$incr]
  set record(last) [expr $last*$incr]
  set record(incr) $incr
  eval "$bar set $total $win $first $last"
}


proc xySet {xx yy} {
  # record xy location of mouse
  global c lastX lastY x y
  set lastX $x
  set lastY $y
  set x [$c canvasx $xx]
  set y [$c canvasy $yy]
}


proc xySetDrag {xx yy} {
  # record xy location of mouse and autoscroll
  global c lastX lastY x y canvx canvy
  xySet $xx $yy
  # auto scrolling
  if {$x < $canvx(first) && $x < $lastX} {
    $c xview [expr "$x/$canvx(incr)"]
    set x [$c canvasx $xx]
  }
  if {($x > $canvx(last)) && ($x > $lastX)} {
    $c xview [expr "($x-$canvx(last)+$canvx(first))/$canvx(incr)"]
    set x [$c canvasx $xx]
  }
  if {$y < $canvy(first) && $y < $lastY} {
    $c yview [expr "$y/$canvy(incr)"]
    set y [$c canvasy $yy]
  }
  if {$y > $canvy(last) && $y > $lastY} {
    $c yview [expr "($y-$canvy(last)+$canvy(first))/$canvy(incr)"]
    set y [$c canvasy $yy]
  }
}








