#
# Copyright (C) 1993 Brian Kahn, Steve O'Keefe, Bob Noel
# Copyright (C) 1993 All Rights Reserved
# Copyright (C) 1993 Thu May 13 1993
#
# Complete copyright is held by each of the authors
# Permission to distribute and use for educational purposes
#
# 

proc displayClean {} {
    # clean display
    global c
    foreach type {bold highlight} {
        foreach obj [$c find withtag $type] {
            itemUn$type $obj
    	}
    }
    foreach obj [$c find withtag boxed] {
    	$c itemconfigure $obj -outline {}
    }
}


proc displayRefresh {} {
    global c warningMessage
    displayClean
    # load the fire set for enabled trans, conflicts
    fireSetLoad
    foreach tag {ConflictPlace EnabledTrans FiredArc BlockedTrans} {
    	$c dtag $tag

    	# remark display
    	global display$tag
        if {[set display$tag]} {
    	    if {[catch gui${tag}Set group]} {
    	    	set warningMessage "Bad ${tag}Set"
    	    } else {
                foreach obj $group {
    		    $c addtag $tag withtag $obj
    	    	    if {$tag=="BlockedTrans"} {
    	    	    	itemHighlight $obj
    	    	    } else {
           	    	itemBold $obj
    	    	    }
    	    	}
    	    }
    	}
    }
    markingRefresh
}


proc itemBold {{tag selected}} {
    # make tag display bold
    global c
    global bold_[itemType $tag]_width
    set item [itemBase $tag]
    $c addtag bold withtag $item
    $c itemconfig $item -width [set bold_[itemType $tag]_width]
}
    


proc itemHighlight {tag} {
    global c
    global high_place_color high_trans_color high_arc_color high_text_color
    global bold_arc_width xpetri_dir
    set item [itemBase $tag]
    $c addtag highlight withtag $item
    case [$c type $item] in {
        {oval} {
    	    $c itemconfig $item \
    	    	-stipple @${xpetri_dir}/bitmap/wide_weave \
      	      	-fill $high_place_color
    	}
        {rectangle} {$c itemconfig $item \
    	    -stipple @${xpetri_dir}/bitmap/wide_weave \
      	    -fill $high_trans_color}
        {line} {$c itemconfig $item \
    	    -stipple @${xpetri_dir}/bitmap/stipple \
      	    -fill $high_arc_color \
    	    -width $bold_arc_width}
        {text} {$c itemconfig $item -stipple {} -fill $high_text_color}

    }
}


proc itemUnbold {{tag selected}} {
    # make tag display bold
    if {[set type [itemType $tag]]==""} {return}
    global c
    global normal_${type}_width
    set item [itemBase $tag]
    $c dtag $item bold
    $c itemconfig $item -width [set normal_${type}_width]
}
    


proc itemUnhighlight {{tag all}} {
    global c
    global normal_place_color normal_trans_color \
    	normal_arc_color normal_text_color \
    	normal_arc_width
    foreach item [itemBase $tag] {
        $c dtag $item highlight
        case [$c type $item] in {
            {oval} {$c itemconfig $item -stipple {} \
    	    	-fill $normal_place_color}
            {rectangle} {$c itemconfig $item -stipple {} \
    	    	-fill $normal_trans_color}
            {line} {$c itemconfig $item -stipple {} \
    	    	-fill $normal_arc_color -width $normal_arc_width}
            {text} {$c itemconfig $item -stipple {} \
    	    	-fill $normal_text_color}
        }
    }
}


proc labelRefresh {} {
    # display all labels
    foreach item [guiPlaceSet] {
    	labelShow $item
    }
    foreach item [guiTransSet] {
    	labelShow $item
    }
}


proc labelShow { item {text {}} } {
    # draw the label for ITEM
    global c label_display
    set id [itemBase $item]
    set type [itemType $item]
    set center [itemCenter $item]
    set x [lindex $center 0]
    set y [lindex $center 1]
    # default to stored label
    if {$text=={}} {
    	set text [guiNodeReadLabel $item]
    }
    # get the trans or place xy sizes
    upvar #0 ${type}X xx ${type}Y yy
    set pos [guiNodeReadLabelOffset $id]
    # give label new location
    set posx [lindex $pos 0]
    set posy [lindex $pos 1]
    case $type in {
    	{trans} {
    	    $c coords labeltext$id \
    	    	[expr $x+($posx*$xx)] [expr $y+($posy*$yy)]
    	    # set label text
    	    if {$label_display(trans)} {
        	$c itemconfigure labeltext$id -text $text
    	    } else {
    	    	$c itemconfigure labeltext$id -text ""
    	    }
    	}
    	{place} {
    	    # adjust the corners closer to circle
    	    if {($posx*$posy)!=0} {
            	$c coords labeltext$id \
    		    [expr $x+($posx*($xx-4))] [expr $y+($posy*($yy-4))]
    	    } else {
    	    	$c coords labeltext$id \
    	    	    [expr $x+($posx*$xx)] [expr $y+($posy*$yy)]
    	    }
    	    # set label text
    	    if {$label_display(place)} {
        	$c itemconfigure labeltext$id -text $text
    	    } else {
    	    	$c itemconfigure labeltext$id -text ""
    	    }
    	}
    }
    # anchor the text in the right "direction"
    labelAnchor $id $pos
}


proc labelAnchor {id pos} {
    # anchor corner of label
    global c
    case $pos in {
    	{{-1 -1}} {$c itemconfigure labeltext$id -anchor se}
    	{{-1 0}} {$c itemconfigure labeltext$id -anchor e}
    	{{-1 1}} {$c itemconfigure labeltext$id -anchor ne}
    	{{0 -1}} {$c itemconfigure labeltext$id -anchor s}
    	{{0 0}} {$c itemconfigure labeltext$id -anchor center}
    	{{0 1}} {$c itemconfigure labeltext$id -anchor n}
    	{{1 -1}} {$c itemconfigure labeltext$id -anchor sw}
    	{{1 0}} {$c itemconfigure labeltext$id -anchor w}
    	{{1 1}} {$c itemconfigure labeltext$id -anchor nw}
    	default {puts "error: labelAnchor '$id' '$pos'"}
    }
}


proc markingRefresh {} {
    # show all markings
    foreach place [guiPlaceSet] {
        markingShow $place [guiPlaceReadMarking $place]
    }
}



proc markingSet {place marking} {
    # set a place's marking
    global c marking_modified
    set marking_modified 1
    set place [itemBase $place]
    guiPlaceWriteMarking $place $marking
    markingShow $place $marking
}



proc markingShow {place marking} {
    global c normal_text_color displayDifference
    set place [itemBase $place]
    set text [lrange [$c itemconfigure markingtext$place -text] 4 4]
    if {$marking!=$text} {
        $c itemconfigure markingtext$place -text $marking 
        if {$displayDifference} {
    	    $c itemconfigure markingbox$place -outline $normal_text_color
    	    $c addtag boxed withtag markingbox$place
    	}
    }
}



proc mk_arc {x y} {
    global c normal_arc_width normal_text_color arcSplinesteps
    # create and return arc 
    # This does all the graphics, not the DM stuff
    # arcs have begin, middle, end
    # endpoints are set by guiArcAdjust after creation
    set item [$c create line $x $y $x $y $x $y \
     	-smooth true -splinesteps $arcSplinesteps \
    	-arrow last -width $normal_arc_width]
    $c addtag arc$item withtag $item
    $c addtag arc withtag $item
    # box provides constant background for text
    $c create oval \
    	[expr $x-6] [expr $y-6] \
    	[expr $x+6] [expr $y+6] \
    	-fill {} -outline {} -width 1 \
    	-tags "arc$item arc weight$item weightbox$item"
    $c create text $x $y \
    	-fill $normal_text_color \
    	-tags "arc$item arc weight$item weighttext$item"
    return $item
}


proc mk_place {x1 y1 x2 y2} {
    global c normal_place_color normal_place_width normal_text_color
    # make a Place
    # This does all the graphics, not the DM stuff
    set item [$c create oval $x1 $y1 $x2 $y2 \
    	-width $normal_place_width -fill $normal_place_color]
    $c addtag place$item withtag $item
    $c addtag place withtag $item
    # box provides constant background for text
    set x [expr ($x1+$x2)/2]
    set y [expr ($y1+$y2)/2]
    $c create oval \
 	[expr $x-10] [expr $y-10] \
    	[expr $x+10] [expr $y+10] \
    	-fill $normal_place_color -outline {} -width 1\
    	-tags "place$item place markingbox$item"
    set center [between $x1 $y1 $x2 $y2]
    $c create text [lindex $center 0] [lindex $center 1] \
    	-fill $normal_text_color \
    	-tags "place$item place markingtext$item"
    $c create text $x1 $y1 \
    	-fill $normal_text_color \
    	-tags "place$item labeltext$item"
    global placeCnt
    incr placeCnt
    return $item
}


proc mk_trans {x1 y1 x2 y2} {
    global c net normal_trans_color normal_trans_width normal_text_color 
    # make a Trans
    # This does all the graphics, not the DM stuff
    set item [$c create rectangle $x1 $y1 $x2 $y2 \
    	-width $normal_trans_width -fill $normal_trans_color]
    $c addtag trans$item withtag $item
    $c addtag trans withtag $item
    global fireStatus
    set fireStatus($item) fire
    # lines are drawn a bit differently, so adjust to fit inside
    set x2 [expr $x2-2]
    set y2 [expr $y2-2]
    # trans decorations - / means hold, X means inhibit
    $c create line $x2 $y1 $x1 $y2 \
    	-fill {} -width $normal_trans_width \
    	-tags "trans$item trans hold$item"
    $c create line $x1 $y1 $x2 $y2 \
    	-fill {} -width $normal_trans_width \
    	-tags "trans$item trans inhibit$item"
    $c create text $x1 $y1 \
    	-fill $normal_text_color -tags "trans$item labeltext$item"
    global transCnt
    incr transCnt
    return $item
}


proc weightShow {arc weight} {
    global c canvasColor
    set arc [itemBase $arc]
    if {$weight==1} then {
        $c itemconfigure weighttext$arc -text ""
        $c itemconfigure weightbox$arc -fill {}
    } else {
        $c itemconfigure weighttext$arc -text $weight
        $c itemconfigure weightbox$arc -fill $canvasColor
    }
}


